# -*- coding: utf-8 -*-
import socket, logging, sys, time
from handler import Handler
from collection import QueueCollection

DEFAULT_HOST = '127.0.0.1'
DEFAULT_PORT = 21122
DEFAULT_PATH = '/tmp/peafowl/'
DEFAULT_TIMEOUT = 60
DEFAULT_PID = '/var/run/peafowl.pid'

class Server(object):
    def __init__(self, *args, **kwargs):
        """
        Initialize a new Peafowl server, but do not accept connections or
        process requests.
        """
        opts = {'host':DEFAULT_HOST, 'port':DEFAULT_PORT, 'path':DEFAULT_PATH, 'timeout':DEFAULT_TIMEOUT, 'debug':False}
        opts.update(kwargs)
        self.opts = opts
        self.queue_collection = QueueCollection(self.opts['path'])
        self.stats = {'bytes_read':0, 'bytes_written':0, 'start_time':time.time(), 'connections':0, 
                      'total_connections':0, 'get_requests':0, 'set_requests':0}
        if opts['debug']:
            logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(levelname)s %(message)s')
    
    def _bind(self):
        try:
            self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.server.bind((self.opts['host'], self.opts['port']))
            self.server.listen(5)
        except socket.error, e:
            if self.server:
                self.server.close()
            logging.error("Could not open socket: %s" % e.message)
            sys.exit(1)
    
    def stop(self):
        if self.server:
            self.server.close()
    
    def start(self):
        """
        Start listening and processing requests.
        """
        self.stats['start_time'] = time.time()
        self._bind()
        logging.debug("Listening to %s on port %s" % (self.opts['host'], self.opts['port']))
        while True:
            try:
                self.stats['connections'] += 1
                client, address = self.server.accept()
                Handler(client, self.queue_collection, self.stats).start()
                self.stats['connections'] -= 1
            except socket.error, e:
                sys.exit(1)




# -*- coding: utf-8 -*-
import time
try:
    import cmemcache as memcache
except ImportError:
    import memcache

WAIT_TIME = 0.25

class PeafowlError(Exception):
    pass
    
class Peafowl(memcache.Client):
    def get(self, *args, **kwargs):
        while True:
            response = super(Peafowl, self).get(*args, **kwargs)
            if response:
                return response
            time.sleep(WAIT_TIME)
    
    def set(self, *args, **kwargs):
        retries = 0
        while retries < 3:
            return_value = super(Peafowl, self).set(*args, **kwargs)
            if not return_value:
                retries += 1
                time.sleep(WAIT_TIME)
            else:
                break;
        else:
            raise PeafowlError("Can't set value")

    def __len__(self, *args, **kwargs):
        statistics = self.get_stats()
        return 1
        




# -*- coding: utf-8 -*-
import os, time, thread
from Queue import Queue
from struct import pack, unpack

SOFT_LOG_MAX_SIZE = 16 * (1024**2) # 16 MB

TRX_CMD_PUSH = "\x00"
TRX_CMD_POP = "\x01"

TRX_PUSH = "\x00%s%s"
TRX_POP = "\x01"

class TransactionLogError(Exception):
    pass

class PersistentQueue(Queue):
    """
    PersistentQueue is a subclass of Python synchronized class. It adds a 
    transactional log to the in-memory Queue, which enables quickly rebuilding
    the Queue in the event of a sever outage.
    """
    def __init__(self, persistence_path, queue_name, debug = False):
        """
        Create a new PersistentQueue at ``persistence_path/queue_name``.
        If a queue log exists at that path, the Queue will be loaded from
        disk before being available for use.
        """
        self.persistence_path = persistence_path
        self.queue_name = queue_name
        self.transaction_lock = thread.allocate_lock()
        self.total_items = 0
        Queue.__init__(self, 0)
        self.initial_bytes = self._replay_transaction_log(debug)
    
    def put(self, value, log = True):
        """
        Pushes ``value`` to the queue. By default, +put+ will write to the
        transactional log. Set ``log`` to ``False`` to override this behaviour.
        """
        if log:
            if not self.transaction_log:
                raise TransactionLogError("No transaction log")
            size = pack("I", len(value))
            self._transaction(TRX_PUSH % (size, value))
        self.total_items += 1
        Queue.put(self, value)
        
    def get(self, log = True):
        """
        Retrieves data from the queue.
        """
        if log and not self.transaction_log:
            raise TransactionLogError("No transaction log")
        value = Queue.get(self, log)
        if log:
            self._transaction("\001")
        return value
    
    def close(self):
        """
        Safely closes the transactional queue.
        """
        self.transaction_lock.acquire()
        not_trx = self.transaction_log
        self.transaction_log = None
        not_trx.close()
    
    def _rotate_log(self):
        self.transaction_log.close()
        os.rename(self._log_path(), "%s.%s" % (self._log_path(), time.time()))
        self._open_log()
    
    def _log_path(self):
        return os.path.join(self.persistence_path, self.queue_name)
    
    def _open_log(self):
        self.transaction_log = os.fdopen(os.open(self._log_path(), os.O_RDWR|os.O_CREAT), "rb+")
        self.log_size = os.path.getsize(self._log_path())
        
    def _replay_transaction_log(self, debug = False):
        self._open_log()
        bytes_read = 0
        
        if debug:
            print "Reading back transaction log for %s" % self.queue_name
    
        while True:
            cmd = self.transaction_log.read(1)
            if not cmd:
                break;
            if cmd == TRX_CMD_PUSH:
                if debug:
                    print ">"
                raw_size = self.transaction_log.read(4)
                size = unpack("I", raw_size)
                data = self.transaction_log.read(size[0])
                if not data:
                    continue
                self.put(data, False)
                bytes_read += len(data)
            elif cmd == TRX_CMD_POP:
                if debug:
                    print "<"
                bytes_read -= len(self.get(False))
            else:
                if debug:
                    print "Error reading transaction log: I don't understand '%s' (skipping)." % cmd
        if debug:
            print "done."
        return bytes_read
        
    def _transaction(self, data):
        if not self.transaction_log:
            raise TransactionLogError("No transaction log")
        
        try:
            self.transaction_lock.acquire()
            self.transaction_log.write(data)
            self.transaction_log.flush()
            self.log_size += len(data)
            if self.log_size > SOFT_LOG_MAX_SIZE and self.qsize() == 0:
                self._rotate_log()
        finally:
            self.transaction_lock.release()




# -*- coding: utf-8 -*-
from optparse import OptionParser
import server, os, sys, errno, signal, logging, string, traceback, time

class Runner(object):
    def __init__(self):
        (options, args) = self.parse_options()
        self.process = ProcessHelper(options.log_file, options.pid_file, options.user, options.group)
        self.server = None
        pid = self.process.is_running()
        if pid:
            sys.stderr.write("There is already a Peafowl process running (pid %s), exiting.\n" % pid)
            sys.exit(1)
        elif not pid:
            sys.stderr.write("Cleaning up stale pidfile at %s.\n" % options.pid_file)
        self.start(options)
        
    @staticmethod
    def run():
        Runner()
        
    def parse_options(self):
        parser = OptionParser()
        parser.add_option("-q", "--queue", action="store", type="string", dest="path", help="path to store Peafowl queue logs", default=server.DEFAULT_PATH)
        parser.add_option("-H", "--host", action="store", type="string", dest="host", help="interface on which to listen", default=server.DEFAULT_HOST)
        parser.add_option("-p", "--port", action="store", type="int", dest="port", help="TCP port on which to listen", default=server.DEFAULT_PORT)
        parser.add_option("-d", action="store_true", dest="daemonize", help="run as a daemon", default=False)
        parser.add_option("-P", "--pid", action="store", type="string", dest="pid_file", help="save pid in PID_FILE when using -d option", default=server.DEFAULT_PID)
        parser.add_option("-u", "--user", action="store", type="int", dest="user", help="user to run as")
        parser.add_option("-g", "--group", action="store", type="int", dest="group", help="group to run as")
        parser.add_option("-l", "--log", action="store", type="string", dest="log_file", help="path to print debugging information")
        parser.add_option("-v", action="count", dest="verbosity", help="increase logging verbosity")
        parser.set_defaults(verbose=True)
        return parser.parse_args()
    
    def start(self, options):
        if options.user:
            os.seteuid(options.user)
        if options.group:
            os.setegid(options.group)
            
        if options.daemonize:
            self.process.daemonize()
        
        self.trap_signals()
        self.process.write_pid_file()
        self.server = server.Server(host=options.host, port=options.port, path=options.path)
        self.server.start()
        self.process.remove_pid_file()
    
    def shutdown(self, signal, frame):
        try:
            print "Shutting down."
            logging.info("Shutting down.")
            if self.server:
                self.server.stop()
        except Exception, e:
            trace = string.join(traceback.format_exception(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2]))
            sys.stderr.write("There was an error shutting down: %s\n%s" % (e, trace))
            sys.exit(70)
    
    def trap_signals(self):
        signal.signal(signal.SIGINT, self.shutdown)
        signal.signal(signal.SIGTERM, self.shutdown)
        
class ProcessHelper(object):
    def __init__(self, log_file = None, pid_file = None, user = None, group = None):
        self.log_file = log_file
        self.pid_file = pid_file
        self.user = user
        self.group= group
    
    def safe_fork(self):
        try:
            return os.fork()
        except OSError, e:
            if e.errno == errno.EWOULDBLOCK:
                time.sleep(5)
                self.safe_fork()
    
    def daemonize(self):
        pid = self.safe_fork()
        if pid > 0:
            sys.exit(0)
        os.chdir("/")
        sess_id = os.setsid()
        os.umask(0)
        pid = self.safe_fork()
        if pid > 0:
            sys.exit(0)
        return sess_id
    
    def is_running(self):
        if not self.pid_file:
            return False 
        try:
            pid_file = open(self.pid_file, 'r')
            pid = int(pid_file.read())
            pid_file.close()
            if pid == 0:
                return False
        except IOError, e:
            return False
        
        try:
            os.kill(pid, 0)
            return pid
        except OSError, e:
            if e.errno == errno.ESRCH:
                return None
            elif e.errno == errno.EPERM:
                return pid
    
    def write_pid_file(self):
        if not self.pid_file:
            return
        pid_file = open(self.pid_file, "w")
        pid_file.write(str(os.getpid()))
        os.chmod(self.pid_file, 0644)
        
    def remove_pid_file(self):
        if not self.pid_file:
            return
        os.remove(self.pid_file)
    
        



# -*- coding: utf-8 -*-
import threading, re, time, os, logging
from resource import getrusage, RUSAGE_SELF
from struct import pack, unpack
import peafowl as peafowl

DATA_PACK_FMT = "!I%sp"

# ERROR responses
ERR_UNKNOWN_COMMAND = "CLIENT_ERROR bad command line format\r\n"

# GET Responses
GET_COMMAND = r'^get (.{1,250})\r\n$'
GET_RESPONSE = "VALUE %s %s %s\r\n%s\r\nEND\r\n"
GET_RESPONSE_EMPTY = "END\r\n"

# SET Responses
SET_COMMAND = r'^set (.{1,250}) ([0-9]+) ([0-9]+) ([0-9]+)\r\n$'
SET_RESPONSE_SUCCESS  = "STORED\r\n"
SET_RESPONSE_FAILURE  = "NOT STORED\r\n"
SET_CLIENT_DATA_ERROR = "CLIENT_ERROR bad data chunk\r\nERROR\r\n"

# STAT Response
STATS_COMMAND = r'stats\r\n$'
STATS_RESPONSE = """STAT pid %d\r
STAT uptime %d\r
STAT time %d\r
STAT version %s\r
STAT rusage_user %0.6f\r
STAT rusage_system %0.6f\r
STAT curr_items %d\r
STAT total_items %d\r
STAT bytes %d\r
STAT curr_connections %d\r
STAT total_connections %d\r
STAT cmd_get %d\r
STAT cmd_set %d\r
STAT get_hits %d\r
STAT get_misses %d\r
STAT bytes_read %d\r
STAT bytes_written %d\r
STAT limit_maxbytes %d\r
%s\nEND\r\n"""
QUEUE_STATS_RESPONSE = """STAT queue_%s_items %d\r
STAT queue_%s_total_items %d\r
STAT queue_%s_logsize %d\r
STAT queue_%s_expired_items %d\r"""

class Handler(threading.Thread):
    """
    This is an internal class used by Peafowl Server to handle the
    MemCache protocol and act as an interface between the Server and the
    QueueCollection.
    """
    def __init__(self, socket, queue_collection, stats):
        threading.Thread.__init__(self)
        self.expiry_stats = {}
        self.stats = stats
        self.queue_collection = queue_collection
        self.socket = socket
        self.file = self.socket.makefile("rb")

    def run(self):
        """
        Process incoming commands from the attached client.
        """
        self.stats['total_connections'] += 1
        while True:
            command = self.file.readline()
            if not command:
                break
            logging.debug("Receiving command : %s" % repr(command))
            self.stats['bytes_read'] += len(command)
            self._process(command)
    
    def _process(self, command):
        m = re.match(SET_COMMAND, command)
        if m:
            logging.debug("Received a SET command")
            self.stats['set_requests'] += 1
            self.set(m.group(1), m.group(2), m.group(3), m.group(4))
            return
        m = re.match(GET_COMMAND, command)
        if m:
            logging.debug("Received a GET command")
            self.stats['get_requests'] += 1
            self.get(m.group(1))
            return
        m = re.match(STATS_COMMAND, command)
        if m:
            logging.debug("Received a STATS command")
            self.get_stats()
            return
        logging.debug("Received unknow command")
        self._respond(ERR_UNKNOWN_COMMAND)
    
    def _respond(self, message, *args):
        response = message % args
        self.stats['bytes_written'] += len(response)
        logging.debug("Sending response : %s" % repr(response))
        self.socket.send(response)
    
    def set(self, key, flags, expiry, length):
        length = int(length)
        data = self.file.read(length)
        data_end = self.file.read(2)
        self.stats['bytes_read'] += (length + 2)
        if data_end == '\r\n' and len(data) == length:
            internal_data = pack(DATA_PACK_FMT % (length + 1), int(expiry), data)
            if self.queue_collection.put(key, internal_data):
                logging.debug("SET command is a success")
                self._respond(SET_RESPONSE_SUCCESS)
            else:
                logging.warning("SET command failed")
                self._respond(SET_RESPONSE_FAILURE)
        else:
            logging.error("SET command failed hard")
            self._respond(SET_CLIENT_DATA_ERROR)
    
    def get(self, key):
        now = time.time()
        data = None
        response = self.queue_collection.take(key)
        while response:
            expiry, data = unpack(DATA_PACK_FMT % (len(response) - 4), response)
            if expiry == 0 or expiry >= now:
                break
            if self.expiry_stats.has_key(key):
                self.expiry_stats[key] += 1
            else:
                self.expiry_stats[key] = 1
            expiry, data = None, None
            response = self.queue_collection.take(key)
        if data:
            logging.debug("GET command respond with value")
            self._respond(GET_RESPONSE, key, 0, len(data), data)
        else:
            logging.debug("GET command response was empty")
            self._respond(GET_RESPONSE_EMPTY)
    
    def get_stats(self):
        self._respond(STATS_RESPONSE,
            os.getpid(), # pid
            time.time() - self.stats['start_time'], # uptime
            time.time(), # time
            '0.1', # peafowl version
            getrusage(RUSAGE_SELF)[0],
            getrusage(RUSAGE_SELF)[1], 
            self.queue_collection.get_stats('current_size'),
            self.queue_collection.get_stats('total_items'),
            self.queue_collection.get_stats('current_bytes'),
            self.stats['connections'],
            self.stats['total_connections'],
            self.stats['get_requests'],
            self.stats['set_requests'],
            self.queue_collection.stats['get_hits'],
            self.queue_collection.stats['get_misses'],
            self.stats['bytes_read'],
            self.stats['bytes_written'],
            0,
            self.queue_stats()    
        )
        
    def queue_stats(self):
        response = ''
        for name in self.queue_collection.get_queues():
            queue = self.queue_collection.get_queues(name)
            if self.expiry_stats.has_key(name):
                expiry_stats = self.expiry_stats[name]
            else:
                expiry_stats = 0
            response += QUEUE_STATS_RESPONSE % (name, queue.qsize(), name, queue.total_items, name, queue.log_size, name, expiry_stats)
        return response

    



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup
#from distutils.core import setup

setup(name='peafowl',
      version='0.1',
      description='A light weight server for reliable distributed message passing.',
      author='Timothee Peignier',
      author_email='tim@tryphon.org',
      url='http://pypi.python.org/pypi/peafowl/',
      packages = ['peafowl'],
      scripts=['bin/peafowl'],
      test_suite='tests'
)



# -*- coding: utf-8 -*-
import os, thread
from queue import PersistentQueue

class QueueCollectionError(Exception):
    pass

class QueueCollection(object):
    def __init__(self, path):
        if not os.path.isdir(path) and os.access(path, os.W_OK):
            raise QueueCollectionError("Queue path '%s' is inacessible" % path) 
        self.shutdown_lock = thread.allocate_lock()
        self.path = path
        self.queues = {}
        self.queue_locks = {}
        self.stats = {'current_bytes':0, 'total_items':0, 'get_misses':0, 'get_hits':0}
    
    def put(self, key, data):
        """
        Puts ``data`` onto the queue named ``key``.
        """
        queue = self.get_queues(key)
        if not queue:
            return None
        self.stats['current_bytes'] += len(data)
        self.stats['total_items'] += 1
        queue.put(data)
        return True
    
    def take(self, key):
        """
        Retrieves data from the queue named ``key``.
        """
        queue = self.get_queues(key)
        if not queue or not queue.qsize():
            self.stats['get_misses'] += 1
            return None
        else:
            self.stats['get_hits'] += 1
        result = queue.get()
        self.stats['current_bytes'] -= len(result)
        return result
    
    def get_queues(self, key = None):
        """
        Returns all active queues.
        """
        if self.shutdown_lock.locked():
            return None
            
        if not key:
            return self.queues
        
        if self.queues.has_key(key):
            return self.queues[key]
        
        if not self.queue_locks.has_key(key):
            self.queue_locks[key] = thread.allocate_lock()
        
        if self.queue_locks[key].locked():
            return None
        else:
            try:
                self.queue_locks[key].acquire()
                if not self.queues.has_key(key):
                    self.queues[key] = PersistentQueue(self.path, key)
                    self.stats['current_bytes'] += self.queues[key].initial_bytes
            finally:
                self.queue_locks[key].release()
        return self.queues[key]
        
    def get_stats(self, name = None):
        """
        Returns statistic ``stat_name`` for the ``QueueCollection``.
        Valid statistics are:
            ``get_misses``    Total number of get requests with empty responses
            ``get_hits``      Total number of get requests that returned data
            ``current_bytes`` Current size in bytes of items in the queues
            ``current_size``  Current number of items across all queues
            ``total_items``   Total number of items stored in queues.
        """
        if not name:
            return self.stats
        elif name == 'current_size':
            return self._current_size()
        else:
            return self.stats[name]
    
    def close(self):
        """
        Safely close all queues.
        """
        self.shutdown_lock.acquire()
        for name, queue in self.queues:
            queue.close()
            del self.queues[name]
    
    def _current_size(self):
        size = 0
        for name in self.queues:
            size += self.queues[name].qsize()
        return size
        
